/*
 * COMPONENT_NAME: ldap.clients
 *
 * ABSTRACT: generic program to modify or add entries using LDAP
 *
 * ORIGINS: 202,27
 *
 */
/*
 * Copyright (c) 1995 Regents of the University of Michigan.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that this notice is preserved and that due credit is given
 * to the University of Michigan at Ann Arbor. The name of the University
 * may not be used to endorse or promote products derived from this
 * software without specific prior written permission. This software
 * is provided ``as is'' without express or implied warranty.
 */

/* ldapchangepwd.c - generic program to modify a user's password */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include <sys/types.h>
#include <sys/stat.h>
#if !defined( WIN32 )
#include <sys/file.h>
#endif
#include <fcntl.h>
#if !defined(WIN32)
#include <unistd.h>
#endif
#include <ldap.h>
#include <ldapssl.h>
#include <ldif.h>

#define OLD_PASSWORD 1
#define NEW_PASSWORD  2

#define DEFHOST "localhost"

#if defined( WIN32 ) || defined( _WIN32 )
#define strcasecmp stricmp
#endif

#define safe_realloc( ptr, size ) ( ptr == NULL ? malloc( size ) : \
      realloc( ptr, size ))

#define LDAPMOD_MAXLINE  4096

#ifdef NEEDPROTOS
static void freepmods( LDAPMod **pmods );
#endif /* NEEDPROTOS */

/* Required for password prompting */
#ifdef _WIN32
#include <conio.h>
#else 
/* termios.h is defined by POSIX */
#include <termios.h>
#endif

/* Arbitrary limit to supplied password length */
#define MAX_SUPPLIED_PW_LENGTH 255
char* getPassword( int type );

const char* ssl_err2string( int );

#ifndef _WIN32
#ifdef _GCC3
#include <errno.h>
#else
extern int errno;
#endif
#endif

/* global variables */
static char *prog;
static char *binddn = NULL;
static char *passwd = (char *) "";
static char *newpasswd = (char *) "";
static char *ldaphost = (char *)DEFHOST;
static char *mech = NULL;
static int hoplimit = 10;
static int ldapport = LDAP_PORT;
static int verbose = 0;
static int contoper = 0;
static int referrals = LDAP_OPT_ON;
static int deref = LDAP_DEREF_NEVER;
static LDAP *ld;
static int ldapversion = LDAP_VERSION3;
static int DebugLevel = 0;
static FILE *fp = NULL;
static char *charset = NULL;
static int keys = 0;
static int tls = 0;
static int ssl = 0;
static int use_FIPS_mode = 0; /* 0=off 1=on */
static char *keyfile = NULL;
static char *keyfile_pw = NULL;
static char *keyfile_label = NULL;
static char *debug_ptr = NULL;

LDAPControl** serverControls = NULL;

/* Digest-MD5 */
static char *digest_username = NULL;
static char *digest_realm = NULL;

/*
** Display usage statement and exit.
*/
void usage() {
   fprintf(stderr, "\nSends modify password requests to an LDAP server.\n");
   fprintf(stderr, "usage:\n");
   fprintf(stderr, "    %s -D binddn -w currentpassword -n newpassword [options]\n", prog);
   fprintf(stderr, "where:\n");
   fprintf(stderr, "      binddn: distinguished name of entry to modify\n");
   fprintf(stderr, "options:\n" );
   fprintf(stderr, "    -C charset      character set name to use, as registered with IANA\n");
   fprintf(stderr, "    -d level        set debug level in LDAP library\n");
   fprintf(stderr, "    -D binddn       bind dn\n");
   fprintf(stderr, "    -G realm        realm used for DIGEST-MD5 bind mechanism\n");
   fprintf(stderr, "    -h host         LDAP server host name\n");
   fprintf(stderr, "    -K keyfile      file to use for keys\n");
   fprintf(stderr, "    -m mechanism    perform SASL bind with the given mechanism\n");
   fprintf(stderr, "    -M              Manage referral objects as normal entries.\n");
   fprintf(stderr, "    -n newpassword  new password or '?' for non-echoed prompt.\n");
   fprintf(stderr, "                    use backslash '\\?' to avoid matching single character filenames (UNIX only)\n");
   fprintf(stderr, "    -N key_name     private key name to use in keyfile\n");
   fprintf(stderr, "    -O maxhops      maximum number of referrals to follow in a sequence\n");
   fprintf(stderr, "    -p port         LDAP server port number\n");
   fprintf(stderr, "    -P key_pw       keyfile password\n");
   fprintf(stderr, "    -R              do not chase referrals\n");
   fprintf(stderr, "    -U username     user name for DIGEST-MD5 bind mechanism\n");
   fprintf(stderr, "    -v              verbose mode\n");
   fprintf(stderr, "    -V version      LDAP protocol version (2 or 3; default is 3)\n");
   fprintf(stderr, "    -w password     current bind password or '?' for non-echoed prompt\n");
   fprintf(stderr, "                    use backslash '\\?' to avoid matching single character filenames (UNIX only)\n");
   fprintf(stderr, "    -x              use FIPS mode processing (SSL/TLS only)\n");
   fprintf(stderr, "    -y proxydn      set proxied id for proxied authorization operation \n");
   fprintf(stderr, "    -Y              use a secure LDAP connection (TLS)\n");
   fprintf(stderr, "    -Z              use a secure LDAP connection (SSL)\n");
   exit(1);
}

/*
**      Print a memory error message and exit.
*/
void mem_error(const char *varName) {
   fprintf(stderr,"%s: Memory allocation error occured.\n", varName);
   exit(1);
}

/* Added function to validate numerical parameters.  */
int is_num(char *s) {
   int rc = 0;
   if (s && (*s == '+' || *s == '-'))
      s++;
   while (s && *s) {
      if (!isdigit(*s))
         return rc;
      s++;
   }
   return 1;
}

/*
** Parse command line arguments.
*/
void parse_arguments(int argc, char **argv) {
   int i, port = 0, rc = LDAP_SUCCESS;
   const char *optpattern = "RMxYZv?h:V:p:D:w:d:n:m:O:K:P:N:C:G:U:y:";
   extern char *optarg;
   extern int optind;
   int manageDsa = LDAP_FALSE;
   char *proxyDN=NULL;
   /* proxied authorization control request */
   int Proxy_control_requested = LDAP_FALSE;
   int stringLen=0;


   fp = stdin;
   while ((i = getopt(argc, argv, optpattern)) != EOF) {
      switch ( i ) {
      case 'V':
         ldapversion = atoi(optarg);
         if ((ldapversion != LDAP_VERSION2) && (ldapversion != LDAP_VERSION3)) {
            fprintf(stderr, "Unsupported version level supplied.\n");
            usage();
         }
         break ;
      case 'O':         /* set maximum referral hop count; validate the number  */
         if (!is_num(optarg)) {              
            fprintf(stderr, "Incorrect hop count.\n");
            usage();
         }
         hoplimit = atoi( optarg );
         if (hoplimit < 0) {
            fprintf(stderr, "Incorrect hop count.\n");
            usage();
         }
         break;
      case 'm':         /* use SASL bind mechanism  */
         if (!(mech = strdup ( optarg )))
            mem_error("mech");
         break;
      case 'h':   /* ldap host */
         if (!(ldaphost = strdup( optarg )))
            mem_error("host");
         break;
      case 'D':   /* bind DN */
         if (!(binddn = strdup( optarg )))
            mem_error("bindDN");
         break;
      case 'w':   /* password */
         if (optarg && optarg[0] == '?') {
            passwd = getPassword(OLD_PASSWORD);
         } else
            if (!(passwd = strdup( optarg )))
               mem_error("oldpassword");
         break;
      case 'n':   /* new password */
         if (optarg && optarg[0] == '?') {
            newpasswd = getPassword(NEW_PASSWORD);
         } else
            if (!(newpasswd = strdup( optarg )))
               mem_error("newpassword");
         break;
      case 'd':  /* debug */
         if (!is_num(optarg)) {    
            fprintf(stderr, "Incorrect debug level.\n");
            usage();
         }
         DebugLevel = atoi(optarg);
         break;
      case 'p':  /* port */
         if (!is_num(optarg)) {    
            fprintf(stderr, "Incorrect ldap port.\n");
            usage();
         }
         ldapport = atoi( optarg );
         if (ldapport <= 0) {      
            fprintf(stderr, "Incorrect ldap port.\n");
            usage();
         }
         port = 1;
         break;
      case 'R':    /* don't automatically chase referrals */
         referrals = LDAP_OPT_OFF;
         break;
      case 'M':   /* manage referral objects as normal entries */
         manageDsa = LDAP_TRUE;
         if ((rc = ldap_add_control (LDAP_MANAGEDSAIT_OID, 0 , NULL, LDAP_OPT_ON, &serverControls)) != LDAP_SUCCESS){
            fprintf( stderr, "ldap_add_control: %d, %s\n",
                     rc, ldap_err2string( rc ) );
            exit(rc);
         }
         break;
      case 'y':   /* Proxied authorization control */
            rc=ldap_proxy_dn_prefix(&proxyDN,optarg);
            if (rc == LDAP_SUCCESS) {
                stringLen=strlen(proxyDN);
                rc = ldap_add_control (LDAP_PROXY_CONTROL_OID,stringLen,
                      proxyDN,1,&serverControls);
               free(proxyDN);
              }
          if (rc != LDAP_SUCCESS) {
              fprintf(stderr, "Incorrect Proxied DN specified. \n");
              usage();
               }
           else  Proxy_control_requested=LDAP_TRUE;
          break;
      case 'v':   /* verbose mode */
         verbose++;
         break;
      case 'K':  /* keyfile */
         if (!(keyfile = strdup( optarg )))
            mem_error("keyfile name");
         keys = 1;
         break;
      case 'P':  /* keyfile password */
         if (!(keyfile_pw = strdup( optarg )))
            mem_error("keyfile password");
         break;
      case 'N':  /* keyfile label*/
         if (!(keyfile_label = strdup( optarg )))
            mem_error("keyfile label");
         break;
      case 'x':  /* enable FIPS mode*/
	 use_FIPS_mode = 1;
         break;
      case 'Y':  /* TLS */
         if (ssl) {
            fprintf(stderr, "Cannot use both SSL and TLS at the same time.\n");
            usage();
         }
         tls = 1;
         break;
      case 'Z':  /* SSL */
         if (tls) {
            fprintf(stderr, "Cannot use both SSL and TLS at the same time.\n");
            usage();
         }
         ssl = 1;
         break;
      case 'C':  /* charset */
         if (!(charset = strdup( optarg )))
            mem_error("charset");
         break;
      case 'U':  /* DIGEST-MD5 Username */
         if (!(digest_username = strdup( optarg )))
            mem_error("Username");
         break;
      case 'G':  /* DIGEST-MD5 Realm */
         if (!(digest_realm = strdup( optarg )))
            mem_error("realm");
         break;
      case '?':
      default:
         usage();
      }
   }

   if ((! (tls || ssl)) && keys) /* Specified a key file, but no security protocol */
      ssl = 1;                  /* Use SSL by default */

   if ( manageDsa && ( ldapversion == LDAP_VERSION2 ) ) {
      fprintf( stderr, "\n-M option requires version 3. -M ignored.\n\n");
      manageDsa = LDAP_FALSE;
   }

   if (argc - optind != 0)
      usage();

   if ((strlen(passwd) < 1) || (strlen(newpasswd) < 1))
      usage();

   if (binddn == NULL)
      usage();

   /* Use default SSL port if none specified */
   if (( port == 0 ) && ( ssl ))
      ldapport = LDAPS_PORT;

   if ( ! DebugLevel )
      if ( ( debug_ptr = getenv ( "LDAP_DEBUG" ) ) )
         DebugLevel = atoi ( debug_ptr );
}

/*  Get a password from the user but don't display it. */
char* getPassword( int type ) {
   char supplied_password[ MAX_SUPPLIED_PW_LENGTH + 1 ]; /* Buffer for password */

#ifdef _WIN32

   char in = '\0';                  /* Input character */
   int len = 0;                    /* Length of password */
#else

   struct termios echo_control;
   struct termios save_control;

   int fd = 0;                 /* File descriptor */
   int attrSet = 0;            /* Checked later for reset */

   /* Get the file descriptor associated with stdin. */
   fd = fileno( stdin );

   if (tcgetattr( fd, &echo_control ) != -1) {
      save_control = echo_control;
      echo_control.c_lflag &= ~( ECHO | ECHONL );

      if (tcsetattr( fd, TCSANOW, &echo_control ) == -1) {
         fprintf(stderr, "Internal error setting terminal attribute.\n");
         exit( errno );
      }

      attrSet = 1;
   }
#endif

   /* Prompt for a password. */
   if (type == OLD_PASSWORD)
      fputs( "Enter Old password ==> ", stdout );
   else
      fputs( "Enter New password ==> ", stdout );
   fflush( stdout );

#ifdef _WIN32
   /* Windows 9x/NT will always read from the console, i.e.,
      piped or redirected input will be ignored. */
   while ( in != '\r' && len <= MAX_SUPPLIED_PW_LENGTH ) {
      in = _getch();

      if (in != '\r') {
         supplied_password[len] = in;
         len++;

      } else {
         supplied_password[len] = '\0';
      }
   }
#else
   /* Get the password from stdin. */
   fgets( supplied_password, MAX_SUPPLIED_PW_LENGTH, stdin );

   /* Remove the newline at the end. */
   supplied_password[strlen( supplied_password ) - 1] = '\0';

#endif

#ifndef _WIN32
   /* Reset the terminal. */
   if (attrSet && tcsetattr( fd, TCSANOW, &save_control ) == -1) {
      fprintf(stderr, "Unable to reset the display.\n");
   }
#endif

   fprintf( stdout, "\n" );

   return ( supplied_password == NULL ) ? supplied_password : strdup( supplied_password );
}

/*
** Rebind callback function.
*/
int rebindproc(LDAP *ld, char **dnp, char **pwp, int *methodp, int freeit) {
   if ( !freeit ) {
      if ( mech != NULL)
         *methodp = LDAP_AUTH_SASL;
      else
         *methodp = LDAP_AUTH_SIMPLE;
      if ( binddn != NULL ) {
         *dnp = strdup( binddn );
         *pwp = strdup ( passwd );
      } else {
         *dnp = NULL;
         *pwp = NULL;
      }
   } else {
      free ( *dnp );
      free ( *pwp );
   }
   return ( LDAP_SUCCESS );
}

/*
** Connect and bind to server.
*/

void connect_to_server() {
   int failureReasonCode, rc, authmethod;
   struct berval ber;
   struct berval *server_creds;
   int msgidp = 0;
   int errcode = 0;
   int error = 0;
   int rc2 = 0;
   LDAPMessage *res;
   int controlres = 0;
   int controlerr = 0;
   int controlwarn = 0;
   char *errmsg = NULL;
   const char *errormsg = NULL;
   const char *errormsgctrl = NULL;
   const char *warnmsgctrl = NULL;
   LDAPControl **returnedControls = NULL;
   LDAPControl **pwdPolicy_controls = NULL;
   LDAPControl **bind_client_controls = NULL;

   /* Set FIPS mode processing if requested for TLS and SSL. */
   if (( tls == 1 || ssl == 1 ) && use_FIPS_mode == 1 ) {
      if (verbose)
	 printf( "ldap_ssl_set_fips_mode_np(1)\n");
      
      rc = ldap_ssl_set_fips_mode_np( 1 ); 
      if (rc != LDAP_SUCCESS) {
	 fprintf( stderr,
		  "ldap_ssl_set_fips_mode_np failed\n" );
	 exit( rc ) ;
      }
   }

   /* call ldap_ssl_client_init if V3 and SSL */
   if ((tls || ssl) && (ldapversion == LDAP_VERSION3)) {
      if ( keyfile == NULL ) {
         keyfile = getenv("SSL_KEYRING");
         if (keyfile != NULL) {
            keyfile = strdup(keyfile);
         }
      }

      if (verbose)
         printf( "ldap_ssl_client_init( %s, %s, 0, &failureReasonCode )\n",
                 ((keyfile) ? keyfile : "NULL"),
                 ((keyfile_pw) ? keyfile_pw : "NULL"));
      rc = ldap_ssl_client_init( keyfile, keyfile_pw, 0, &failureReasonCode ) ;
      if (rc != LDAP_SUCCESS) {
         fprintf( stderr,
                  "ldap_ssl_client_init failed! rc == %d, failureReasonCode == %d\n",
                  rc, failureReasonCode );
         fprintf( stderr, "%s\n", ssl_err2string( failureReasonCode ));
         exit( 1 ) ;
      }
   }

   /* open connection to server */
   if (ldapversion == LDAP_VERSION3) {
      if (ssl) {
         if (verbose)
            printf("ldap_ssl_init( %s, %d, %s )\n",
                   ((ldaphost) ? ldaphost : "NULL"), ldapport,
                   ((keyfile_label) ? keyfile_label : "NULL"));
         ld = ldap_ssl_init( ldaphost, ldapport, keyfile_label ) ;
         if (ld == NULL) {
            fprintf( stderr, "ldap_ssl_init failed\n" ) ;
            perror( "ldap_ssl_init" ) ;
            exit( 1 ) ;
         }
      } else {
         if (verbose)
            printf("ldap_init(%s, %d) \n",
                   ((ldaphost) ? ldaphost : "NULL"), ldapport);
         if ((ld = ldap_init(ldaphost, ldapport)) == NULL) {
            perror("ldap_init");
            exit(1);
         }
      }
   } else {
      if ( verbose )
         printf("ldap_open( %s, %d )\n",
                ((ldaphost) ? ldaphost : "NULL"), ldapport );
      if (( ld = ldap_open( ldaphost, ldapport )) == NULL ) {
         perror( "ldap_open" );
         exit( 1 );
      }
   }

   /* set options */
   ldap_set_option (ld, LDAP_OPT_PROTOCOL_VERSION, (void * )&ldapversion);
   if (ldapversion == LDAP_VERSION3) {
      ldap_set_option (ld, LDAP_OPT_DEBUG, (void * )&DebugLevel);
      ldap_set_option (ld, LDAP_OPT_DEREF, (void * )&deref);
      ldap_set_option( ld, LDAP_OPT_REFHOPLIMIT, (void *)&hoplimit);
      ldap_set_option( ld, LDAP_OPT_REBIND_FN, (void *)rebindproc);
   } else {
      ldap_set_option (ld, LDAP_OPT_DEBUG, (void * )(long)DebugLevel);
      ldap_set_option (ld, LDAP_OPT_DEREF, (void * )(long)deref);
      ldap_set_option( ld, LDAP_OPT_REFHOPLIMIT, (void *)(long)hoplimit);
   }
   /* don't chase referrals during the bind */
   ldap_set_option (ld, LDAP_OPT_REFERRALS, (void * )LDAP_OPT_OFF);
   if (binddn != NULL)
      ldap_set_rebind_proc( ld, (LDAPRebindProc)rebindproc );
   if (charset != NULL) {
      if (ldap_set_iconv_local_charset(charset) != LDAP_SUCCESS) {
         fprintf(stderr, "unsupported charset %s\n", charset);
         exit(0);
      }
      ldap_set_option(ld, LDAP_OPT_UTF8_IO, (void *)LDAP_UTF8_XLATE_ON);
   }

   /* call ldap_ssl_start if V2 and SSL */
   if ( (tls || ssl) && (ldapversion == LDAP_VERSION2)) {
      if ( keyfile == NULL ) {
         keyfile = getenv("SSL_KEYRING");
         if (keyfile != NULL) {
            keyfile = strdup(keyfile);
         }
      }
      if (verbose)
         printf( "ldap_ssl_start( ld, %s, %s, %s )\n",
                 ((keyfile) ? keyfile : "NULL"),
                 ((keyfile_pw) ? keyfile_pw : "NULL"),
                 ((keyfile_label) ? keyfile_label : "NULL"));

      rc = ldap_ssl_start(ld, keyfile, keyfile_pw, keyfile_label);
      if ( rc != 0) {
         fprintf( stderr, "ldap_ssl_start failed! RC = %d\n", rc);
         errormsg = ldap_err2string(ldap_get_errno(ld));
         fprintf(stderr,"ldap_ssl_start: %s \n", errormsg);
         exit( 1 );
      }
   }

   if (tls) {                   /* Start the TLS session */
      if ((rc = ldap_start_tls_s_np(ld, NULL)) != LDAP_SUCCESS) {
         fprintf(stderr, "ldap_start_tls_s() failed! RC = %d\n", rc);
         errormsg = ldap_err2string(ldap_get_errno(ld));
         fprintf(stderr,"ldap_start_tls_s_np: %s\n", errormsg);
         exit(rc);
      }
   }

   /* bind to server */
   if (ldapversion == LDAP_VERSION3) {

      /*Set password policy control*/
      if ((rc = ldap_add_control (LDAP_PWDPOLICY_CONTROL_OID, 0 , NULL, LDAP_OPT_OFF, &pwdPolicy_controls)) != LDAP_SUCCESS){
         fprintf( stderr, "ldap_add_control: %d, %s\n",
                  rc, ldap_err2string( rc ) );
         exit(rc);
      }

      if ( ! mech ) /* Use simple bind */
      {
         ber.bv_len = strlen ( passwd );
         ber.bv_val = passwd;
         rc = ldap_sasl_bind(ld, binddn, LDAP_SASL_SIMPLE, &ber, pwdPolicy_controls, NULL, &msgidp);
         rc2 = rc; /* Store this result code for later user. */
         if (rc == LDAP_SUCCESS) {
            error = ldap_result( ld, msgidp, 1, (struct timeval *) NULL, &res );
            rc = ldap_parse_result( ld, res, &errcode, NULL, &errmsg,
                                    NULL, &returnedControls, 1 );
         }
      }
      else /* Presence of mechanism means SASL bind */ {
         /*
         ** Special case for mech="EXTERNAL".  Unconditionally set bind DN
         ** and credentials to NULL.  This option should be used in tandem
         ** with SSL and client authentication.  For other SASL mechanisms,
         ** use the specified bind DN and credentials.
         */
         if (strcmp(mech, LDAP_MECHANISM_EXTERNAL) == 0) {
            rc = ldap_sasl_bind_s (ld, NULL, mech, NULL, NULL, NULL, &server_creds);
            if (rc != LDAP_SUCCESS ) {
               errormsg = ldap_err2string(ldap_get_errno(ld));
               fprintf(stderr,"ldap_sasl_bind_s: %s \n", errormsg);
               fprintf(stderr,"ldap_sasl_bind_s: extended error code: %i\n", ldap_get_exterror(ld));
               exit ( rc );
            }
         } else if (strcmp(mech, LDAP_MECHANISM_GSSAPI) == 0) {
            rc = ldap_sasl_bind_s (ld, NULL, mech, NULL, pwdPolicy_controls, NULL, &server_creds);
            if (rc != LDAP_SUCCESS ) {
               errormsg = ldap_err2string(ldap_get_errno(ld));
               fprintf(stderr,"ldap_sasl_bind_s: %s \n", errormsg);
               ldap_get_option( ld, LDAP_OPT_EXT_GSS_ERR, &failureReasonCode);
               fprintf(stderr,"ldap_sasl_bind_s: extended error code: Major Status: %i\n", failureReasonCode);
               fprintf(stderr,"ldap_sasl_bind_s: extended error code: Minor Status: %i\n", ldap_get_exterror(ld));
               exit ( rc );
            }
         } else /* other SASL mechanisms */ {
            if (strcmp(mech,LDAP_MECHANISM_DIGEST_MD5) == 0) {
               if (digest_username) {
                  if ((rc = ldap_add_control (IBM_CLIENT_DIGEST_USER_NAME_OID, strlen(digest_username) , digest_username, LDAP_TRUE, &bind_client_controls)) != LDAP_SUCCESS) {
                     ldap_perror (ld, "ldap_add_control");
                     exit ( rc );
                  }
               }
               if (digest_realm) {
                  if ((rc = ldap_add_control (IBM_CLIENT_DIGEST_REALM_NAME_OID, strlen(digest_realm) , digest_realm, LDAP_TRUE, &bind_client_controls)) != LDAP_SUCCESS) {
                     ldap_perror (ld, "ldap_add_control");
                     exit ( rc );
                  }
               }
            }

            ber.bv_len = strlen ( passwd );
            ber.bv_val = passwd;
            rc = ldap_sasl_bind_s (ld, binddn, mech, &ber, pwdPolicy_controls,
                                      bind_client_controls, &server_creds);
            if (rc != LDAP_SUCCESS ) {
               errormsg = ldap_err2string(ldap_get_errno(ld));
               fprintf(stderr,"ldap_sasl_bind_s: %s \n", errormsg);
               fprintf(stderr,"ldap_sasl_bind_s: extended error code: %i\n", ldap_get_exterror(ld));
               exit ( rc );
            }
         }

         /* Get any controls sent by the server during the sasl_bind_s */
         rc2 = rc;

         if ( rc == LDAP_SUCCESS ) {
            rc = ldap_get_bind_controls( ld, &returnedControls );
         }
      }

      /*Free the client controls*/
      ldap_controls_free(bind_client_controls);

      if (rc == LDAP_SUCCESS) {
         if (returnedControls != NULL) {
            ldap_parse_pwdpolicy_response(returnedControls, &controlerr, &controlwarn, &controlres);
         }
      }
      if (rc2 != LDAP_SUCCESS) {
         errormsg = ldap_err2string(rc2);
      } else if (rc != LDAP_SUCCESS) {
         errormsg = ldap_err2string(rc);
      } else if (errcode != LDAP_SUCCESS) {
         errormsg = ldap_err2string(errcode);
      }
      if (controlerr) {
         errormsgctrl = ldap_pwdpolicy_err2string(controlerr);
      }
      if (controlwarn) {
         warnmsgctrl = ldap_pwdpolicy_err2string(controlwarn);
      }
      if ((errormsgctrl != NULL) || (warnmsgctrl != NULL)) {
         fprintf(stderr,"ldap_simple_bind:  ");
         if (errormsg != NULL) {
            fprintf(stderr,errormsg);
            fprintf(stderr," --- ");
         }
         if (errormsgctrl != NULL) {
            fprintf(stderr,errormsgctrl);
            fprintf(stderr,"\n");
         }
         if (warnmsgctrl != NULL) {
            fprintf(stderr,warnmsgctrl, controlres);
            fprintf(stderr,"\n");
         }
      } else if (errormsg != NULL) {
         fprintf(stderr,"ldap_simple_bind:   %s \n", errormsg);
      }
      if (rc2 != LDAP_SUCCESS) {
         /*
         ** ldap_get_exterror only valuable for ssl communication.
         ** In this example, for LDAP v3, the bind is the first
         ** instance in which communication actually flows to the
         ** server.  So, if there is an ssl configuration error or
         ** other ssl problem, this will be the first instance where
         ** it will be detected.
         */
         if (ssl) {
            fprintf( stderr, "Attempted communication over SSL.\n");
            fprintf( stderr, "  The extended error is %d.\n", ldap_get_exterror(ld));
         }
         exit( rc2 );
      } else if (errcode != LDAP_SUCCESS) {
         exit( errcode );
      }

   } else /* LDAP V2 */ {
      authmethod = LDAP_AUTH_SIMPLE;

      rc = ldap_bind_s( ld, binddn, passwd, authmethod );

      if ( rc != LDAP_SUCCESS ) {
         errormsg = ldap_err2string(ldap_get_errno(ld));
         fprintf(stderr,"ldap_bind_s: %s \n", errormsg);
         exit( rc );
      }
   }

   /* during the operation, use user setting to chase referrals */
   ldap_set_option (ld, LDAP_OPT_REFERRALS, (void * )(long)referrals);

   /* Free the password policy controls. */
   ldap_controls_free(pwdPolicy_controls);
}

/* ssl_err2string returns a string describing the error associated with
** the given error code from an SSL related call.
*/
const char * ssl_err2string( int ssl_error ) {
   switch ( ssl_error ) {
   case LDAP_SSL_EXTENDED_OK:
      return LDAP_SSL_EXTENDED_OK_STR;
   case LDAP_SSL_EXTENDED_INVALID_HANDLE:
      return LDAP_SSL_EXTENDED_INVALID_HANDLE_STR;
   case LDAP_SSL_EXTENDED_API_NOT_AVAILABLE:
      return LDAP_SSL_EXTENDED_API_NOT_AVAILABLE_STR;
   case LDAP_SSL_EXTENDED_INTERNAL_ERROR:
      return LDAP_SSL_EXTENDED_INTERNAL_ERROR_STR;
   case LDAP_SSL_EXTENDED_INSUFFICIENT_STORAGE:
      return LDAP_SSL_EXTENDED_INSUFFICIENT_STORAGE_STR;
   case LDAP_SSL_EXTENDED_INVALID_STATE:
      return LDAP_SSL_EXTENDED_INVALID_STATE_STR;
   case LDAP_SSL_EXTENDED_KEY_LABEL_NOT_FOUND:
      return LDAP_SSL_EXTENDED_KEY_LABEL_NOT_FOUND_STR;
   case LDAP_SSL_EXTENDED_CERTIFICATE_NOT_AVAILABLE:
      return LDAP_SSL_EXTENDED_CERTIFICATE_NOT_AVAILABLE_STR;
   case LDAP_SSL_EXTENDED_ERROR_CERT_VALIDATION:
      return LDAP_SSL_EXTENDED_ERROR_CERT_VALIDATION_STR;
   case LDAP_SSL_EXTENDED_ERROR_CRYPTO:
      return LDAP_SSL_EXTENDED_ERROR_CRYPTO_STR;
   case LDAP_SSL_EXTENDED_ERROR_ASN:
      return LDAP_SSL_EXTENDED_ERROR_ASN_STR;
   case LDAP_SSL_EXTENDED_ERROR_LDAP:
      return LDAP_SSL_EXTENDED_ERROR_LDAP_STR;
   case LDAP_SSL_EXTENDED_ERROR_UNKNOWN_ERROR:
      return LDAP_SSL_EXTENDED_ERROR_UNKNOWN_ERROR_STR;
   case LDAP_SSL_EXTENDED_OPEN_CIPHER_ERROR:
      return LDAP_SSL_EXTENDED_OPEN_CIPHER_ERROR_STR;
   case LDAP_SSL_EXTENDED_KEYFILE_IO_ERROR:
      return LDAP_SSL_EXTENDED_KEYFILE_IO_ERROR_STR;
   case LDAP_SSL_EXTENDED_KEYFILE_INVALID_FORMAT:
      return LDAP_SSL_EXTENDED_KEYFILE_INVALID_FORMAT_STR;
   case LDAP_SSL_EXTENDED_KEYFILE_DUPLICATE_KEY:
      return LDAP_SSL_EXTENDED_KEYFILE_DUPLICATE_KEY_STR;
   case LDAP_SSL_EXTENDED_KEYFILE_DUPLICATE_LABEL:
      return LDAP_SSL_EXTENDED_KEYFILE_DUPLICATE_LABEL_STR;
   case LDAP_SSL_EXTENDED_BAD_FORMAT_OR_INVALID_PASSWORD:
      return LDAP_SSL_EXTENDED_BAD_FORMAT_OR_INVALID_PASSWORD_STR;
   case LDAP_SSL_EXTENDED_KEYFILE_CERT_EXPIRED:
      return LDAP_SSL_EXTENDED_KEYFILE_CERT_EXPIRED_STR;
   case LDAP_SSL_EXTENDED_ERROR_LOAD_LDAP_SSL_EXTENDED_LIB:
      return LDAP_SSL_EXTENDED_ERROR_LOAD_LDAP_SSL_EXTENDED_LIB_STR;
   case LDAP_SSL_EXTENDED_PENDING_CLOSE_ERROR:
      return LDAP_SSL_EXTENDED_PENDING_CLOSE_ERROR_STR;
   case LDAP_SSL_EXTENDED_NO_KEYFILE_PASSWORD:
      return LDAP_SSL_EXTENDED_NO_KEYFILE_PASSWORD_STR;
   case LDAP_SSL_EXTENDED_KEYRING_OPEN_ERROR:
      return LDAP_SSL_EXTENDED_KEYRING_OPEN_ERROR_STR;
   case LDAP_SSL_EXTENDED_RSA_TEMP_KEY_PAIR:
      return LDAP_SSL_EXTENDED_RSA_TEMP_KEY_PAIR_STR;
   case LDAP_SSL_EXTENDED_ERROR_LDAP_NO_SUCH_OBJECT:
      return LDAP_SSL_EXTENDED_ERROR_LDAP_NO_SUCH_OBJECT_STR;
   case LDAP_SSL_EXTENDED_ERROR_LDAP_INVALID_CREDENTIALS:
      return LDAP_SSL_EXTENDED_ERROR_LDAP_INVALID_CREDENTIALS_STR;
   case LDAP_SSL_EXTENDED_ERROR_BAD_INDEX:
      return LDAP_SSL_EXTENDED_ERROR_BAD_INDEX_STR;
   case LDAP_SSL_EXTENDED_CLOSE_FAILED:
      return LDAP_SSL_EXTENDED_CLOSE_FAILED_STR;
   case LDAP_SSL_EXTENDED_ERROR_BAD_DATE:
      return LDAP_SSL_EXTENDED_ERROR_BAD_DATE_STR;
   case LDAP_SSL_EXTENDED_ERROR_NO_CIPHERS:
      return LDAP_SSL_EXTENDED_ERROR_NO_CIPHERS_STR;
   case LDAP_SSL_EXTENDED_ERROR_NO_CERTIFICATE:
      return LDAP_SSL_EXTENDED_ERROR_NO_CERTIFICATE_STR;
   case LDAP_SSL_EXTENDED_ERROR_BAD_CERTIFICATE:
      return LDAP_SSL_EXTENDED_ERROR_BAD_CERTIFICATE_STR;
   case LDAP_SSL_EXTENDED_ERROR_UNSUPPORTED_CERTIFICATE_TYPE:
      return LDAP_SSL_EXTENDED_ERROR_UNSUPPORTED_CERTIFICATE_TYPE_STR;
   case LDAP_SSL_EXTENDED_ERROR_IO:
      return LDAP_SSL_EXTENDED_ERROR_IO_STR;
   case LDAP_SSL_EXTENDED_ERROR_BAD_KEYFILE_LABEL:
      return LDAP_SSL_EXTENDED_ERROR_BAD_KEYFILE_LABEL_STR;
   case LDAP_SSL_EXTENDED_ERROR_BAD_KEYFILE_PASSWORD:
      return LDAP_SSL_EXTENDED_ERROR_BAD_KEYFILE_PASSWORD_STR;
   case LDAP_SSL_EXTENDED_ERROR_BAD_KEY_LEN_FOR_EXPORT:
      return LDAP_SSL_EXTENDED_ERROR_BAD_KEY_LEN_FOR_EXPORT_STR;
   case LDAP_SSL_EXTENDED_ERROR_BAD_MESSAGE:
      return LDAP_SSL_EXTENDED_ERROR_BAD_MESSAGE_STR;
   case LDAP_SSL_EXTENDED_ERROR_BAD_MAC:
      return LDAP_SSL_EXTENDED_ERROR_BAD_MAC_STR;
   case LDAP_SSL_EXTENDED_ERROR_UNSUPPORTED:
      return LDAP_SSL_EXTENDED_ERROR_UNSUPPORTED_STR;
   case LDAP_SSL_EXTENDED_ERROR_BAD_CERT_SIG:
      return LDAP_SSL_EXTENDED_ERROR_BAD_CERT_SIG_STR;
   case LDAP_SSL_EXTENDED_ERROR_BAD_CERT:
      return LDAP_SSL_EXTENDED_ERROR_BAD_CERT_STR;
   case LDAP_SSL_EXTENDED_ERROR_BAD_PEER:
      return LDAP_SSL_EXTENDED_ERROR_BAD_PEER_STR;
   case LDAP_SSL_EXTENDED_ERROR_PERMISSION_DENIED:
      return LDAP_SSL_EXTENDED_ERROR_PERMISSION_DENIED_STR;
   case LDAP_SSL_EXTENDED_ERROR_SELF_SIGNED:
      return LDAP_SSL_EXTENDED_ERROR_SELF_SIGNED_STR;
   case LDAP_SSL_EXTENDED_ERROR_NO_READ_FUNCTION:
      return LDAP_SSL_EXTENDED_ERROR_NO_READ_FUNCTION_STR;
   case LDAP_SSL_EXTENDED_ERROR_NO_WRITE_FUNCTION:
      return LDAP_SSL_EXTENDED_ERROR_NO_WRITE_FUNCTION_STR;
   case LDAP_SSL_EXTENDED_ERROR_SOCKET_CLOSED:
      return LDAP_SSL_EXTENDED_ERROR_SOCKET_CLOSED_STR;
   case LDAP_SSL_EXTENDED_ERROR_BAD_V2_CIPHER:
      return LDAP_SSL_EXTENDED_ERROR_BAD_V2_CIPHER_STR;
   case LDAP_SSL_EXTENDED_ERROR_BAD_V3_CIPHER:
      return LDAP_SSL_EXTENDED_ERROR_BAD_V3_CIPHER_STR;
   case LDAP_SSL_EXTENDED_ERROR_BAD_SEC_TYPE:
      return LDAP_SSL_EXTENDED_ERROR_BAD_SEC_TYPE_STR;
   case LDAP_SSL_EXTENDED_ERROR_BAD_SEC_TYPE_COMBINATION:
      return LDAP_SSL_EXTENDED_ERROR_BAD_SEC_TYPE_COMBINATION_STR;
   case LDAP_SSL_EXTENDED_ERROR_HANDLE_CREATION_FAILED:
      return LDAP_SSL_EXTENDED_ERROR_HANDLE_CREATION_FAILED_STR;
   case LDAP_SSL_EXTENDED_ERROR_INITIALIZATION_FAILED:
      return LDAP_SSL_EXTENDED_ERROR_INITIALIZATION_FAILED_STR;
   case LDAP_SSL_EXTENDED_ERROR_LDAP_NOT_AVAILABLE:
      return LDAP_SSL_EXTENDED_ERROR_LDAP_NOT_AVAILABLE_STR;
   case LDAP_SSL_EXTENDED_ERROR_NO_PRIVATE_KEY:
      return LDAP_SSL_EXTENDED_ERROR_NO_PRIVATE_KEY_STR;
   case LDAP_SSL_EXTENDED_ERROR_PKCS11_LIBRARY_NOTLOADED:
      return LDAP_SSL_EXTENDED_ERROR_PKCS11_LIBRARY_NOTLOADED_STR;
   case LDAP_SSL_EXTENDED_ERROR_PKCS11_TOKEN_LABELMISMATH:
      return LDAP_SSL_EXTENDED_ERROR_PKCS11_TOKEN_LABELMISMATH_STR;
   case LDAP_SSL_EXTENDED_ERROR_PKCS11_TOKEN_NOTPRESENT:
      return LDAP_SSL_EXTENDED_ERROR_PKCS11_TOKEN_NOTPRESENT_STR;
   case LDAP_SSL_EXTENDED_ERROR_PKCS11_TOKEN_BADPASSWORD:
      return LDAP_SSL_EXTENDED_ERROR_PKCS11_TOKEN_BADPASSWORD_STR;
   case LDAP_SSL_EXTENDED_ERROR_INVALID_V2_HEADER:
      return LDAP_SSL_EXTENDED_ERROR_INVALID_V2_HEADER_STR;
   case LDAP_SSL_EXTENDED_INVALID_BUFFER_SIZE:
      return LDAP_SSL_EXTENDED_INVALID_BUFFER_SIZE_STR;
   case LDAP_SSL_EXTENDED_WOULD_BLOCK:
      return LDAP_SSL_EXTENDED_WOULD_BLOCK_STR;
   case LDAP_SSL_EXTENDED_ERROR_NOT_SSLV3:
      return LDAP_SSL_EXTENDED_ERROR_NOT_SSLV3_STR;
   case LDAP_SSL_EXTENDED_MISC_INVALID_ID:
      return LDAP_SSL_EXTENDED_MISC_INVALID_ID_STR;
   case LDAP_SSL_EXTENDED_ATTRIBUTE_INVALID_ID:
      return LDAP_SSL_EXTENDED_ATTRIBUTE_INVALID_ID_STR;
   case LDAP_SSL_EXTENDED_ATTRIBUTE_INVALID_LENGTH:
      return LDAP_SSL_EXTENDED_ATTRIBUTE_INVALID_LENGTH_STR;
   case LDAP_SSL_EXTENDED_ATTRIBUTE_INVALID_ENUMERATION:
      return LDAP_SSL_EXTENDED_ATTRIBUTE_INVALID_ENUMERATION_STR;
   case LDAP_SSL_EXTENDED_ATTRIBUTE_INVALID_SID_CACHE:
      return LDAP_SSL_EXTENDED_ATTRIBUTE_INVALID_SID_CACHE_STR;
   case LDAP_SSL_EXTENDED_ATTRIBUTE_INVALID_NUMERIC_VALUE:
      return LDAP_SSL_EXTENDED_ATTRIBUTE_INVALID_NUMERIC_VALUE_STR;
   case LDAP_SSL_EXTENDED_CONFLICTING_VALIDATION_SETTING:
      return LDAP_SSL_EXTENDED_CONFLICTING_VALIDATION_SETTING_STR;
   default:
      return "Unknown SSL error";
   }
}


/*
** Add operation to the modify structure.
*/
void addmodifyop(LDAPMod ***pmodsp, int modop, const char *attr, char *value, int vlen) {
   LDAPMod * *pmods;
   int i, j;
   struct berval *bvp;

   /*
   ** Data can be treated as binary (wire ready) if one of the
   ** following applies:
   **     1) it was base64 encode
   **     2) charset is not defined  
   **     3) read from an external file
   */

   i = 0;
   pmods = *pmodsp;
   if ( pmods != NULL ) {
      for ( ; pmods[ i ] != NULL; ++i ) {
         if ( strcasecmp( pmods[ i ]->mod_type, attr ) == 0 &&
              pmods[ i ]->mod_op == modop ) {
            break;
         }
      }
   }

   if ( pmods == NULL || pmods[ i ] == NULL ) {
      if (( pmods = (LDAPMod * *)safe_realloc( pmods, (i + 2) *
                                               sizeof( LDAPMod * ))) == NULL ) {
         perror( "safe_realloc" );
         exit( 1 );
      }
      *pmodsp = pmods;
      pmods[ i + 1 ] = NULL;
      if (( pmods[ i ] = (LDAPMod * )calloc( 1, sizeof( LDAPMod ))) == NULL ) {
         perror( "calloc" );
         exit( 1 );
      }
      pmods[ i ]->mod_op = modop;
      if (( pmods[ i ]->mod_type = strdup( attr )) == NULL ) {
         perror( "strdup" );
         exit( 1 );
      }
   }

   if ( value != NULL ) {
      if (modop & LDAP_MOD_BVALUES) {
         j = 0;
         if ( pmods[ i ]->mod_bvalues != NULL ) {
            for ( ; pmods[ i ]->mod_bvalues[ j ] != NULL; ++j ) {
               ;
            }
         }
         if (( pmods[ i ]->mod_bvalues =
               (struct berval **)safe_realloc( pmods[ i ]->mod_bvalues,
                                               (j + 2) * sizeof( struct berval *))) == NULL ) {
            perror( "safe_realloc" );
            exit( 1 );
         }
         pmods[ i ]->mod_bvalues[ j + 1 ] = NULL;
         if (( bvp = (struct berval *)malloc( sizeof( struct berval )))
             == NULL ) {
            perror( "malloc" );
            exit( 1 );
         }
         pmods[ i ]->mod_bvalues[ j ] = bvp;

         bvp->bv_len = vlen;
         if (( bvp->bv_val = (char *)malloc( vlen + 1 )) == NULL ) {
            perror( "malloc" );
            exit( 1 );
         }
         memmove( bvp->bv_val, value, vlen );
         bvp->bv_val[ vlen ] = '\0';
      } else {
         j = 0;
         if ( pmods[ i ]->mod_values != NULL ) {
            for ( ; pmods[ i ]->mod_values[ j ] != NULL; ++j ) {
               ;
            }
         }
         if (( pmods[ i ]->mod_values =
               (char **)safe_realloc( pmods[ i ]->mod_values,
                                      (j + 2) * sizeof( char *))) == NULL ) {
            perror( "safe_realloc" );
            exit( 1 );
         }
         pmods[ i ]->mod_values[ j + 1 ] = NULL;
         if (( pmods[ i ]->mod_values[ j ] = strdup( value )) == NULL) {
            perror( "strdup" );
            exit( 1 );
         }
      }
   }
}


/*
** Print a binary value. If charset is not specified then check to
** see if string is printable anyway.
*/
void print_binary(struct berval *bval) {
   int i, binary;

   printf( "\tBINARY (%ld bytes) ", bval->bv_len);
   if (charset == NULL) {
      binary = 0;
      for (i = 0; (i < bval->bv_len) && (!binary); ++i)
         if (!isprint(bval->bv_val[i]))
            binary = 1;
      if (!binary)
         for (i = 0; i < bval->bv_len; ++i)
            putchar(bval->bv_val[i]);
   }
   putchar('\n');
}


/*
** Modify or add an entry.
*/
int dopasswordmodify( char *dn, LDAPMod **pmods) {
   int i, j, op, rc, rc2;
   struct berval *bvp;
   int errcode = 0;
   int error = 0;
   int msgid = 0;
   LDAPMessage *res;
   int controlres = 0;
   int controlerr = 0;
   int controlwarn = 0;
   char *errmsg = NULL;
   const char *errormsg = NULL;
   const char *errormsgctrl = NULL;
   const char *warnmsgctrl = NULL;
   LDAPControl **returnedControls = NULL;


   if ( pmods == NULL ) {
      fprintf( stderr, "%s: no attributes to change or add (entry %s)\n",
               prog, dn );
      return ( LDAP_PARAM_ERROR );
   }

   printf( "changing password for entry %s\n", dn );

   if ( verbose ) {
      for ( i = 0; pmods[ i ] != NULL; ++i ) {
         op = pmods[ i ]->mod_op & ~LDAP_MOD_BVALUES;
         printf( "%s %s:\n", op == LDAP_MOD_REPLACE ?
                 "replace" : op == LDAP_MOD_ADD ?
                 "add" : "delete", pmods[ i ]->mod_type );
         if (pmods[i]->mod_op & LDAP_MOD_BVALUES) {
            if (pmods[ i ]->mod_bvalues != NULL) {
               for (j = 0; pmods[i]->mod_bvalues[j] != NULL; ++j)
                  print_binary(pmods[i]->mod_bvalues[j]);
            }
         } else {
            if (pmods[i]->mod_values != NULL) {
               for (j = 0; pmods[i]->mod_values[j] != NULL; ++j)
                  printf("\t%s\n", pmods[i]->mod_values[j]);
            }
         }
      }
   }

   if ((rc = ldap_add_control (LDAP_PWDPOLICY_CONTROL_OID, 0 , NULL, LDAP_OPT_OFF, &serverControls)) != LDAP_SUCCESS){
      fprintf( stderr, "ldap_add_control: %d, %s\n",
               rc, ldap_err2string( rc ) );
   }

   rc = ldap_modify_ext( ld, dn, pmods, serverControls,
                         NULL, &msgid);

   rc2 = rc;
   if (rc == LDAP_SUCCESS) {
      error = ldap_result( ld, msgid, 1, (struct timeval *) NULL, &res );
      rc = ldap_parse_result( ld, res, &errcode, NULL, &errmsg,
                              NULL, &returnedControls, 1 );
   }
   if (rc == LDAP_SUCCESS) {
      if (returnedControls != NULL) {
         ldap_parse_pwdpolicy_response(returnedControls, &controlerr, &controlwarn, &controlres);
      }
   }
   if (rc2 != LDAP_SUCCESS) {
      errormsg = ldap_err2string(rc2);
   } else if (rc != LDAP_SUCCESS) {
      errormsg = ldap_err2string(rc);
   } else if (errcode != LDAP_SUCCESS) {
      errormsg = ldap_err2string(errcode);
   }
   if (controlerr) {
      errormsgctrl = ldap_pwdpolicy_err2string(controlerr);
   }
   if (controlwarn) {
      warnmsgctrl = ldap_pwdpolicy_err2string(controlwarn);
   }
   if ((errormsgctrl != NULL) || (warnmsgctrl != NULL)) {
      if (errormsg != NULL) {
         fprintf(stderr,errormsg);
         fprintf(stderr," --- ");
      }
      if (errormsgctrl != NULL) {
         fprintf(stderr,errormsgctrl);
         fprintf(stderr,"\n");
      }
      if (warnmsgctrl != NULL) {
         fprintf(stderr,warnmsgctrl, controlres);
         fprintf(stderr,"\n");
      }
   } else if (errormsg != NULL) {
      fprintf(stderr," %s \n", errormsg);
   }
   if ( verbose ) {
      printf( "ldapchangepwd complete\n" );
   }
   putchar( '\n' );

   /* Return the proper error code */
   if (rc2)
	   return (rc2);
   else if (rc)
	   return (rc);
   else
	   return (errcode);
}


int main( int argc, char **argv ) {
   char *p, buf[ 4096 ];
   int rc, i, authmethod, len, havedn;
   LDAPMod **pmods = NULL;
   FILE * cf_fd;
   const char *type = "userpassword";

   /* Strip off any path info on program name */
#if defined( WIN32 ) || defined( _WIN32 )

   if ((prog = strrchr(argv[0], '\\')) != NULL)
      ++prog;
   else
      prog = argv[0];
#else

   if (prog = strrchr(argv[0], '/'))
      ++prog;
   else
      prog = argv[0];
#endif

   /* parse command line arguments */
   parse_arguments(argc, argv);


   /* connect to server */
   connect_to_server();

   rc = 0;
   addmodifyop(&pmods, LDAP_MOD_DELETE, type, passwd, strlen(passwd));
   addmodifyop(&pmods, LDAP_MOD_ADD, type, newpasswd, strlen(passwd));
   rc = dopasswordmodify(binddn, pmods);


   if ( tls )
      ldap_stop_tls_s_np ( ld );
   else 
      /* unbind and exit */
      ldap_unbind(ld);

   if(serverControls);
   ldap_controls_free(serverControls);
   exit(rc);
}

